/* eslint-env jest */
const ModECS = require('../index')

describe('ModECS', () => {

    let engine

    beforeEach(() => {
        engine = ModECS()
        engine.registerComponent('POSITION', {x:0,y:0})
    })

    it('should create a component', () => {
        expect(() => engine.createComponent('POSITION')).not.toThrow()
    })

    it('should create a component with values', () => {
        engine.compile()

        const position = engine.createComponent('POSITION', {x:2,y:2})

        expect(position.x).toEqual(2)
        expect(position.y).toEqual(2)
    })

    it('should create an entity', () => {
        const id = engine.createEntity()
        expect(id).toBe(0)
    })

    it('should add registered component to an entity', () => {
        const id = engine.createEntity()
        engine.addEntity(id)
        expect(() => engine.addComponent(id, 'POSITION')).not.toThrow()
    })
})

describe('Scenario 1', () => {
    const engine = ModECS()

    engine.registerComponent('POSITION', {x:0,y:0})
    engine.registerComponent('TARGET', {x:0,y:0})


    engine.registerSystem(
        'Move',
        ['POSITION','TARGET'],
        () => (position, target, id) => {
            position.x += target.x
            position.y += target.y
        }
    )

    // engine.compile()

    const e1 = engine.createEntity()
    engine.addEntity(e1)
    engine.addComponent(e1, 'POSITION')
    engine.addComponent(e1, 'TARGET')
    
    const e2 = engine.createEntity()
    engine.addEntity(e2)
    engine.addComponent(e2, 'POSITION')
    engine.addComponent(e2, 'TARGET')


    it('should', () => {
        const moveView = engine.createView('POSITION')

        engine.removeComponent(e1, 'POSITION')

        expect(moveView.POSITION).toHaveLength(2)

        engine.updateComponent(e2, 'POSITION', {x:1,y:1})
        engine.updateComponent(e2, 'TARGET', {x:11,y:11})


        const moveView2 = engine.createView('POSITION', 'TARGET')

        engine.process()

        expect(moveView2.POSITION).toHaveLength(1)

        

    })

})

describe('Scenario 2', () => {
    const engine = ModECS()

    engine.registerComponent('POSITION', {x:0,y:0})
    engine.registerComponent('TARGET', {x:0,y:0})
    engine.registerComponent('ROTATION', {angle:0})
    engine.registerComponent('OTHER', {x:0})


    engine.registerSystem(
        'Move',
        ['POSITION','TARGET'],
        () => (position, target, id) => {
            position.x += target.x
            position.y += target.y
        }
    )

    const e1 = engine.createEntity()
    engine.addEntity(e1)

    engine.addComponent(e1, 'POSITION')
    engine.addComponent(e1, 'TARGET')
    engine.addComponent(e1, 'ROTATION')

    const e2 = engine.createEntity()
    engine.addEntity(e2)

    engine.addComponent(e2, 'POSITION')
    engine.addComponent(e2, 'TARGET')
    engine.addComponent(e2, 'OTHER')


    it('should', () => {
        const moveView = engine.createView('POSITION')
        const targetView = engine.createView('TARGET')

        const customView = engine.createView('POSITION', 'TARGET', 'OTHER')
        const targetRotationView = engine.createView('TARGET','ROTATION')


        const e3 = engine.createEntity()
        engine.addEntity(e3)
        engine.addComponent(e3, 'OTHER')
        engine.addComponent(e3, 'TARGET')
        engine.addComponent(e3, 'OTHER', {x:4})

        
        expect(moveView.POSITION).toHaveLength(2)
        expect(targetView.TARGET).toHaveLength(3)

        expect(customView.POSITION).toHaveLength(1)
        expect(customView.TARGET).toHaveLength(1)
        expect(customView.OTHER).toHaveLength(1)
        
        const e4 = engine.createEntity()
        engine.addEntity(e4)
        engine.addComponent(e4, 'POSITION')
        engine.addComponent(e4, 'TARGET')
        engine.addComponent(e4, 'OTHER', {x:4})

        
        expect(customView.POSITION).toHaveLength(2)
        expect(customView.TARGET).toHaveLength(2)
        expect(customView.OTHER).toHaveLength(2)

        

    })

})

describe('Scenario 3', () => { 
    const engine = ModECS()

    engine.registerComponent('POSITION', {x:0,y:0})
    engine.registerComponent('TARGET', {x:1,y:1})

    engine.registerSystem(
        'Move',
        ['POSITION','TARGET'],
        () => (position, target, id) => {
            position.x += target.x
            position.y += target.y
        }
    )

    // engine.compile()

    const e1 = engine.createEntity()
    engine.addEntity(e1)
    engine.addComponent(e1, 'POSITION')
    engine.addComponent(e1, 'TARGET')

    const e2 = engine.createEntity()
    engine.addEntity(e2)
    engine.addComponent(e2, 'POSITION')
    engine.addComponent(e2, 'TARGET')


    it('should', () => {
        const moveView = engine.createView('POSITION')

        expect(moveView.POSITION).toHaveLength(2)

        engine.removeComponent(e1, 'POSITION')

        engine.process()

        expect(moveView.POSITION).toHaveLength(1)

        engine.updateComponent(e2, 'POSITION', {x:1,y:1})
        engine.updateComponent(e2, 'TARGET', {x:11,y:11})

        const moveView2 = engine.createView('POSITION', 'TARGET')

        expect(moveView2.POSITION).toHaveLength(1)

        const snapshot = engine.snapshot()

        const engine2 = ModECS({snapshot})

        console.log(engine2.snapshot())

    })

})
